# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
Heavy lifting geometry for IDF surfaces.

PyClipper is used for clipping.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import MutableSequence
from itertools import product

from eppy.geometry.surface import area
from geomeppy.segments import Segment
from geomeppy.transformations import align_face
from geomeppy.transformations import invert_align_face
from geomeppy.utilities import almostequal
from geomeppy.vectors import Vector2D
from geomeppy.vectors import Vector3D
from geomeppy.vectors import normalise_vector
from shapely import wkt

import numpy as np
import pyclipper as pc


class Polygon(MutableSequence):
    """Two-dimensional polygon."""
    n_dims = 2

    def __init__(self, vertices):
        super(Polygon, self).__init__()
        self.vertices = [Vector2D(*v) for v in vertices]
    
    def __repr__(self):
        class_name = type(self).__name__
        return '{}({!r})'.format(class_name, self.vertices)

    def __len__(self):
        return len(self.vertices)
            
    def __delitem__(self, key):
        del self.vertices[key]
    
    def __getitem__(self, key):
        return self.vertices[key]
    
    def __setitem__(self, key, value):
        self.vertices[key] = value

    def __eq__(self, other):
        if self.__dict__ == other.__dict__:  # try the simple case first
            return True
        else:  # also cover same shape in different rotation
            if self.difference(other):
                return False
            if self.normal_vector == other.normal_vector:
                return True
        return False
    
    def __add__(self, other):
        if len(self) == len(other) and hasattr(other[0], '__len__'):
            # add together two equal polygons
            vertices = [v1 + v2 for v1, v2 in zip(self, other)]
        elif len(self[0]) == len(other):
            # translate by a vector
            vertices = [v + other for v in self]
        else:
            raise ValueError("Incompatible objects: %s + %s" % (self, other))
        return self.__class__(vertices)
    
    def __sub__(self, other):
        if len(self) == len(other) and hasattr(other[0], '__len__'):
            # add together two equal polygons
            vertices = [v1 - v2 for v1, v2 in zip(self, other)]
        elif len(self[0]) == len(other):
            # translate by a vector
            vertices = [v - other for v in self]
        else:
            raise ValueError("Incompatible objects: %s + %s" % (self, other))
        return self.__class__(vertices)
    
    def from_wkt(self, wkt_poly):
        """Convert a wkt representation of a polygon to GeomEppy.
        
        This also accounts for the possible presence of inner rings by linking
        them to the outer ring.
        
        Parameters
        ----------
        wkt_poly : str
            A text representation of a polygon in well known text (wkt) format.
        
        Returns
        -------
        Polygon3D
        
        """
        poly = wkt.loads(wkt_poly)
        exterior = Polygon3D(poly.exterior.coords)
        if poly.interiors:
            # make the exterior into a geomeppy poly
            for inner_ring in poly.interiors:
                # make the interior into a geomeppy poly
                interior = Polygon3D(inner_ring.coords)
                # find the nearest points on the exterior and interior
                links = product(interior, exterior)
                links = sorted(links, key=lambda x: distance(x[0], x[1]))
                on_interior = links[0][0]
                on_exterior = links[0][1]
                # join them up
                exterior = Polygon3D(exterior[exterior.index(on_exterior):] + 
                                     exterior[:exterior.index(on_exterior) + 1])
                interior = Polygon3D(interior[interior.index(on_interior):] + 
                                     interior[:interior.index(on_interior) + 1])
                exterior = Polygon3D(exterior[:] + interior[:])
        
        return exterior
    
    @property
    def normal_vector(self):
        as_3d = Polygon3D((v.x, v.y, 0) for v in self)
        return as_3d.normal_vector
    
    @property
    def bounding_box(self):
        aligned = align_face(self)
        top_left = Vector3D(min(aligned.xs), max(aligned.ys), max(aligned.zs))
        bottom_left = Vector3D(min(aligned.xs), min(aligned.ys), min(aligned.zs))
        bottom_right = Vector3D(max(aligned.xs), min(aligned.ys), min(aligned.zs))
        top_right = Vector3D(max(aligned.xs), max(aligned.ys), max(aligned.zs))
        
        bbox = Polygon3D([top_left, bottom_left, bottom_right, top_right])
        return invert_align_face(self, bbox)
        
    def insert(self, key, value):
        self.vertices.insert(key, value)
                
    @property
    def points_matrix(self):
        """[[x1, x2,... xn]
            [y1, y2,... yn]
            [z1, z2,... zn]  # all 0 for 2D polygon
        """
        points = np.zeros((len(self.vertices), self.n_dims))
        for i, v in enumerate(self.vertices):
            points[i,:] = pt_to_array(v, dims=self.n_dims)
        return points
    
    @property
    def edges(self):
        """A list of edges represented as Segment objects.
        """
        vertices = self.vertices
        edges = [Segment(vertices[i], vertices[(i+1) % len(self)])
                for i in range(len(self))]
        return edges
    
    @property
    def xs(self):
        return [pt.x for pt in self.vertices]

    @property
    def ys(self):
        return [pt.y for pt in self.vertices]
   
    @property
    def zs(self):
        return [0.0] * len(self.vertices)
   
    @property
    def vertices_list(self):
        """A list of the vertices in the format required by pyclipper.
        
        Returns
        -------
        list of tuples
            Like [(x1, y1), (x2, y2),... (xn, yn)].
        
        """
        return [pt_to_tuple(pt, dims=self.n_dims) for pt in self.vertices]
    
    @property
    def area(self):
        return area(self)
    
    def invert_orientation(self):
        """Reverse the order of the vertices.
        
        This is to create a matching surface, e.g. the other side of a wall.
        
        Returns
        -------
        Polygon3D
        
        """
        return self.__class__(reversed(self.vertices))
        
    def project_to_3D(self, example3d):
        """Project the 2D polygon rotated into 3D space.
        
        This is used to return a previously rotated 3D polygon back to its
        original orientation, or to to put polygons generated from pyclipper
        into the desired orientation.
        
        Parameters
        ----------
        example3D : Polygon3D
            A 3D polygon in the desired plane.
                
        Returns
        -------
        Polygon3D
        
        """
        points = self.points_matrix
        proj_axis = example3d.projection_axis
        a = example3d.distance
        v = example3d.normal_vector
        projected_points = project_to_3D(points, proj_axis, a, v)
        return Polygon3D(projected_points)
    
    def union(self, poly):
        """Union with another 2D polygon.
        
        Parameters
        ----------
        poly : Polygon
            The clip polygon.

        Returns
        -------
        list
        
        """
        return union_2D_polys(self, poly)
        
    def intersect(self, poly):
        """Intersect with another 2D polygon.
        
        Parameters
        ----------
        poly : Polygon
            The clip polygon.

        Returns
        -------
        list or False
            False if no intersection, otherwise a list of lists of 2D
            coordinates representing each intersection.
        
        """
        return intersect_2D_polys(self, poly)
        
    def difference(self, poly):
        """Intersect with another 2D polygon.
        
        Parameters
        ----------
        poly : Polygon
            The clip polygon.

        Returns
        -------
        list or False
            False if no intersection, otherwise a list of lists of 2D
            coordinates representing each difference.
        
        """
        return difference_2D_polys(self, poly)

def distance(pt1, pt2):
    direction = pt1 - pt2
    return sum(x ** 2 for x in direction)
    
    
def break_polygons(poly, hole):
    """Break up a surface with a hole in it.
    
    This produces two surfaces, neither of which have a hole in them.
    
    Parameters
    ----------
    poly : Polygon3D
        The surface with a hole in.
    hole : Polygon3D
        The hole.
    
    Returns
    -------
    list
        Two Polygon3D objects.
        
    """
    # take the two closest points on the surface perimeter
    links = product(poly, hole)
    links = sorted(links, key=lambda x: distance(x[0], x[1]))

    first_on_poly = links[0][0]
    last_on_poly = links[1][0]

    first_on_hole = links[1][1]
    last_on_hole = links[0][1]

    coords = poly[:] + poly[:]  # a double loop
    section_on_poly = []
    for item in coords:
        if item == first_on_poly:
            section_on_poly.append(item)
        elif section_on_poly:
            section_on_poly.append(item)
            if item == last_on_poly:
                break

    coords = reversed(hole[:] + hole[:])  # a double loop
    section_on_hole = []
    for item in coords:
        if item == first_on_hole:
            section_on_hole.append(item)
        elif section_on_hole:
            section_on_hole.append(item)
            if item == last_on_hole:
                break    

    new_poly = section_on_poly + section_on_hole
        
    new_poly = Polygon3D(new_poly)
    union = hole.union(new_poly)
    union = union[0]
    new_poly2 = poly.difference(union)[0]
    if not almostequal(new_poly.normal_vector, poly.normal_vector):
        print("inverting 1")
        new_poly = new_poly.invert_orientation()
    if not almostequal(new_poly2.normal_vector, poly.normal_vector):
        print("inverting 2")
        new_poly2 = new_poly2.invert_orientation()
#    view_polygons({'blue': [new_poly], 'red': [hole]})
#    view_polygons({'blue': [new_poly, new_poly2], 'red': [hole]})
    
    return [new_poly, new_poly2]
    
    
def prep_2D_polys(poly1, poly2):
    """Prepare two 2D polygons for clipping operations.
    
    Parameters
    ----------
    poly1 : Polygon
        The subject polygon.
    poly2 : Polygon
        The clip polygon.
    
    Returns
    -------
    Pyclipper object
    
    """
    s1 = pc.scale_to_clipper(poly1.vertices_list)
    s2 = pc.scale_to_clipper(poly2.vertices_list)
    clipper = pc.Pyclipper()
    clipper.AddPath(s1, poly_type=pc.PT_SUBJECT, closed=True)
    clipper.AddPath(s2, poly_type=pc.PT_CLIP, closed=True)
    return clipper


def union_2D_polys(poly1, poly2):
    """Union of two 2D polygons.
    
    Find the combined shape of poly1 and poly2.
    
    Parameters
    ----------
    poly1 : Polygon
        The subject polygon.
    poly2 : Polygon
        The clip polygon.
    
    Returns
    -------
    list or False
        False if no intersection, otherwise a list of lists of 2D coordinates
        representing each intersection.
        
    """
    clipper = prep_2D_polys(poly1, poly2)        
    intersections = clipper.Execute(
        pc.CT_UNION, pc.PFT_NONZERO, pc.PFT_NONZERO)
    polys = process_clipped_2D_polys(intersections)
    results = []
    for poly in polys:
        if poly.normal_vector == poly1.normal_vector:
            results.append(poly)
        else:
            results.append(poly.invert_orientation())

    return results


def intersect_2D_polys(poly1, poly2):
    """Intersect two 2D polygons.
    
    Find the area/s that poly1 shares with poly2.
    
    Parameters
    ----------
    poly1 : Polygon
        The subject polygon.
    poly2 : Polygon
        The clip polygon.
    
    Returns
    -------
    list or False
        False if no intersection, otherwise a list of lists of 2D coordinates
        representing each intersection.
        
    """
    clipper = prep_2D_polys(poly1, poly2)        
    intersections = clipper.Execute(
        pc.CT_INTERSECTION, pc.PFT_NONZERO, pc.PFT_NONZERO)
    polys = process_clipped_2D_polys(intersections)
    results = []
    for poly in polys:
        if poly.normal_vector == poly1.normal_vector:
            results.append(poly)
        else:
            results.append(poly.invert_orientation())

    return results


def difference_2D_polys(poly1, poly2):
    """Difference from two 2D polygons.
    
    Equivalent to subtracting poly2 from poly1.
    
    Parameters
    ----------
    poly1 : Polygon
        The subject polygon.
    poly2 : Polygon
        The clip polygon.
    
    Returns
    -------
    list or False
        False if no difference, otherwise a list of lists of 2D coordinates
        representing each difference.
        
    """
    clipper = prep_2D_polys(poly1, poly2)        
    differences = clipper.Execute(
        pc.CT_DIFFERENCE, pc.PFT_NONZERO, pc.PFT_NONZERO)
    polys = process_clipped_2D_polys(differences)
    results = []
    for poly in polys:
        if poly.normal_vector == poly1.normal_vector:
            results.append(poly)
        else:
            results.append(poly.invert_orientation())

    return results


def process_clipped_2D_polys(results):
    """Process and return the results of a clipping operation.
    
    Parameters
    ----------
    results : list
        A list of results, potentially empty if the operation found no
        interactions between polygons.
        
    Returns
    -------
    list or False
        False if no intersection, otherwise a list of lists of 2D coordinates
        representing each intersection.
        
    """
    if results:
        results = [pc.scale_from_clipper(r) for r in results]
        return [Polygon(r) for r in results]
    else:
        return []
        

class Polygon3D(Polygon):
    """Three-dimensional polygon."""
    n_dims = 3

    def __init__(self, vertices):
        try:
            self.vertices = [Vector3D(*v) for v in vertices]
        except TypeError:
            self.vertices = vertices

    def __eq__(self, other):
        # check they're in the same plane
        if not almostequal(self.normal_vector, other.normal_vector):
            return False
        if not almostequal(self.distance, other.distance):
            return False
        # if they are in the same plane, check they completely overlap in 2D
        return (self.project_to_2D() == other.project_to_2D())
    
    @property
    def zs(self):
        return [pt.z for pt in self.vertices]

    @property
    def normal_vector(self):
        """Vector perpendicular to the polygon in the outward direction.
        
        Uses Newell's Method.
        
        Returns
        -------
        Vector3D

        """
        return Vector3D(*normal_vector(self.vertices))

    @property
    def distance(self):
        """
        A number where v[0] * x + v[1] * y + v[2] * z = a is the equation of
        the plane containing the polygon (where v is the polygon normal vector).
        
        Returns
        -------
        float
        
        """
        v = self.normal_vector
        pt = self.points_matrix[0]  # arbitrary point in the polygon
        d = np.dot(v, pt)
        return d
    
    @property
    def projection_axis(self):
        """Return an axis which will not lead to degenerate surface.
        
        Returns
        -------
        int
        
        """
        proj_axis = max(range(3), key=lambda i: abs(self.normal_vector[i]))
        return proj_axis
    
    @property
    def is_horizontal(self):
        """Check if polygon is in the xy plane.
        
        Returns
        -------
        bool
        
        """
        return np.array(self.zs).std() < 1e-8
    
    def is_clockwise(self, viewpoint):
        """Check if vertices are ordered clockwise
        
        This function checks the vertices as seen from the viewpoint.
        
        Parameters
        ----------
        viewpoint : Vector3D
        
        Returns
        -------
        bool
        
        """
        arbitrary_pt = self.vertices[0]
        v = arbitrary_pt - viewpoint
        n = self.normal_vector
        sign = np.dot(v, n)
        return sign > 0
    
    def is_coplanar(self, other):
        """Check if polygon is in the same plane as another polygon.
        
        This includes the same plane but opposite orientation.
        
        Parameters
        ----------
        other : Polygon3D
            Another polygon.
        
        Returns
        -------
        bool
        
        """
        n1 = self.normal_vector
        n2 = other.normal_vector
        d1 = self.distance
        d2 = other.distance
        
        if (almostequal(n1, n2) and almostequal(d1, d2)):
            return True
        elif (almostequal(n1, -n2) and almostequal(d1, -d2)):
            return True
        else:
            return False

    @property
    def centroid(self):
        """The centroid of a polygon.
        
        Returns
        -------
        Vector3D
        
        """
        return Vector3D(
            sum(self.xs) / len(self),
            sum(self.ys) / len(self),
            sum(self.zs) / len(self))
    
    def outside_point(self, entry_direction='counterclockwise'):
        """Return a point outside the zone to which the surface belongs.
        
        The point will be outside the zone, respecting the global geometry rules
        for vertex entry direction.
        
        Parameters
        ----------
        entry_direction : str
            Either "clockwise" or "counterclockwise", as seen from outside the
            space.
        
        Returns
        -------
        Vector3D
        
        """
        entry_direction = entry_direction.lower()
        if entry_direction == 'clockwise':
            inside = self.vertices[0] - self.normal_vector
        elif entry_direction == 'counterclockwise':
            inside = self.vertices[0] + self.normal_vector
        else:
            raise ValueError("invalid value for entry_direction '%s'" % 
                             entry_direction)
    
        return inside
    
    def order_points(self, starting_position):
        """Reorder the vertices based on a starting position rule.
        
        Parameters
        ----------
        starting_position : str
            The string that defines vertex starting position in EnergyPlus.
        
        Returns
        -------
        Polygon3D
        
        """
        if starting_position == 'upperleftcorner':
            bbox_corner = self.bounding_box[0]
        elif starting_position == 'lowerleftcorner':
            bbox_corner = self.bounding_box[1]
        elif starting_position == 'lowerrightcorner':
            bbox_corner = self.bounding_box[2]
        elif starting_position == 'upperrightcorner':
            bbox_corner = self.bounding_box[3]
        else:
            raise ValueError(
                '%s is not a valid starting position' % starting_position)
        start_index = self.index(bbox_corner.closest(self))
        new_vertices = [self[(start_index + i) % len(self)]
                        for i in range(len(self))]

        return Polygon3D(new_vertices)
    
    def project_to_2D(self):
        """Project the 3D polygon into 2D space.
        
        This is so that we can perform operations on it using pyclipper library.
        
        Project onto either the xy, yz, or xz plane. (We choose the one that
        avoids degenerate configurations, which is the purpose of proj_axis.)
        
        Returns
        -------
        Polygon3D
        
        """        
        points = self.points_matrix
        projected_points = project_to_2D(points, self.projection_axis)
        
        return Polygon([pt[:2] for pt in projected_points])
    
    def union(self, poly):
        """Union with another 3D polygon.
        
        Parameters
        ----------
        poly : Polygon3D
            The clip polygon.

        Returns
        -------
        list or False
            False if no union, otherwise a list of lists of Polygon3D
            objects representing each union.
        
        """
        return union_3D_polys(self, poly)
        
    def intersect(self, poly):
        """Intersect with another 3D polygon.
        
        Parameters
        ----------
        poly : Polygon3D
            The clip polygon.

        Returns
        -------
        list or False
            False if no intersection, otherwise a list of lists of Polygon3D
            objects representing each intersection.
        
        """
        return intersect_3D_polys(self, poly)
        
    def difference(self, poly):
        """Difference from another 3D polygon.
        
        Parameters
        ----------
        poly : Polygon3D
            The clip polygon.

        Returns
        -------
        list or False
            False if no difference, otherwise a list of lists of Polygon3D
            objects representing each intersection.
        
        """
        return difference_3D_polys(self, poly)
    
    def normalize_coords(self, ggr):
        """Order points, respecting the global geometry rules
        
        Parameters
        ----------
        ggr : EpPBunch
            GlobalGeometryRules object.
        
        Returns
        -------
        Polygon3D
        
        """
        try:
            entry_direction = ggr.Vertex_Entry_Direction
        except AttributeError:
            entry_direction = 'counterclockwise'            
        outside_point = self.outside_point(entry_direction)
        return normalize_coords(self, outside_point, ggr)

    
def normal_vector(poly):
    """Return the unit normal vector of a polygon.
    
    We use Newell's Method since the cross-product of two edge vectors is not
    valid for concave polygons.
    https://www.opengl.org/wiki/Calculating_a_Surface_Normal#Newell.27s_Method
    
    Parameters
    ----------
    
    """
    n = [0.0, 0.0, 0.0]

    for i, v_curr in enumerate(poly):
        v_next = poly[(i+1) % len(poly)]
        n[0] += (v_curr.y - v_next.y) * (v_curr.z + v_next.z)
        n[1] += (v_curr.z - v_next.z) * (v_curr.x + v_next.x)
        n[2] += (v_curr.x - v_next.x) * (v_curr.y + v_next.y)
    
    return normalise_vector(n)


def prep_3D_polys(poly1, poly2):
    """Prepare two 3D polygons for clipping operations.
    
    Parameters
    ----------
    poly1 : Polygon3D
        The subject polygon.
    poly2 : Polygon3D
        The clip polygon.
    
    Returns
    -------
    Pyclipper object
    
    """
    if not poly1.is_coplanar(poly2):
        return False
    poly1 = poly1.project_to_2D()
    poly2 = poly2.project_to_2D()

    s1 = pc.scale_to_clipper(poly1.vertices_list)
    s2 = pc.scale_to_clipper(poly2.vertices_list)
    clipper = pc.Pyclipper()
    clipper.AddPath(s1, poly_type=pc.PT_SUBJECT, closed=True)
    clipper.AddPath(s2, poly_type=pc.PT_CLIP, closed=True)
    
    return clipper


def union_3D_polys(poly1, poly2):
    """Union of two 3D polygons.
    
    Parameters
    ----------
    poly1 : Polygon3D
        The subject polygon.
    poly2 : Polygon3D
        The clip polygon.

    Returns
    -------
    list or False
        A list of lists of Polygon3D objects representing each union.
    
    """
    clipper = prep_3D_polys(poly1, poly2)
    if not clipper:
        return []
    unions = clipper.Execute(
        pc.CT_UNION, pc.PFT_NONZERO, pc.PFT_NONZERO)
    
    polys = process_clipped_3D_polys(unions, poly1)
    
    # orient to match poly1
    results = []
    for poly in polys:
        if almostequal(poly.normal_vector, poly1.normal_vector):
            results.append(poly)
        else:
            results.append(poly.invert_orientation())

    return results


def intersect_3D_polys(poly1, poly2):    
    """Intersection of two 3D polygons.
    
    Parameters
    ----------
    poly1 : Polygon3D
        The subject polygon.
    poly2 : Polygon3D
        The clip polygon.

    Returns
    -------
    list or False
        False if no intersection, otherwise a list of lists of Polygon3D
        objects representing each intersection.
    
    """
    clipper = prep_3D_polys(poly1, poly2)    
    if not clipper:
        return []
    intersections = clipper.Execute(
        pc.CT_INTERSECTION, pc.PFT_NONZERO, pc.PFT_NONZERO)
    
    polys = process_clipped_3D_polys(intersections, poly1)
    # orient to match poly1
    results = []
    for poly in polys:
        if almostequal(poly.normal_vector, poly1.normal_vector):
            results.append(poly)
        else:
            results.append(poly.invert_orientation())

    return results


def difference_3D_polys(poly1, poly2):
    """Difference between two 3D polygons.
    
    Parameters
    ----------
    poly1 : Polygon3D
        The subject polygon.
    poly2 : Polygon3D
        The clip polygon.

    Returns
    -------
    list or False
        False if no difference, otherwise a list of lists of Polygon3D
        objects representing each difference.
    
    """
    clipper = prep_3D_polys(poly1, poly2)
    if not clipper:
        return []
    differences = clipper.Execute(
        pc.CT_DIFFERENCE, pc.PFT_NONZERO, pc.PFT_NONZERO)
    
    polys = process_clipped_3D_polys(differences, poly1)

    # orient to match poly1
    results = []
    for poly in polys:
        if almostequal(poly.normal_vector, poly1.normal_vector):
            results.append(poly)
        else:
            results.append(poly.invert_orientation())

    return results


def process_clipped_3D_polys(results, example3d):
    """Convert 2D clipping results back to 3D.
    
    Parameters
    ----------
    example3d : Polygon3D
        Used to find the plane to project the 2D polygons into.
    
    Returns
    -------
    list or False
        List of Poygon3D if result found, otherwise False.
        
    """
    if results:
        res_vertices = [pc.scale_from_clipper(r) for r in results]
        return [Polygon(v).project_to_3D(example3d) for v in res_vertices]
    else:
        return []

        
def project_to_2D(vertices, proj_axis):
    """Project a 3D polygon into 2D space.
    
    Parameters
    ----------
    vertices : list 
        The three-dimensional vertices of the polygon.
    proj_axis : int
        The axis to project into.
    a : float
        Distance to the origin for the plane to project into.
    v : list
        Normal vector of the plane to project into.
    
    Returns
    -------
    list
        The transformed vertices.
    
    """
    points = [project(x, proj_axis) for x in vertices]
    return points


def project(pt, proj_axis):
    """Project point pt onto either the xy, yz, or xz plane
    
    We choose the one that avoids degenerate configurations, which is the
    purpose of proj_axis.
    See http://stackoverflow.com/a/39008641/1706564
    
    """
    return tuple(c for i, c in enumerate(pt) if i != proj_axis)


def project_to_3D(vertices, proj_axis, a, v):
    """Project a 2D polygon into 3D space.
    
    Parameters
    ----------
    vertices : list 
        The two-dimensional vertices of the polygon.
    proj_axis : int
        The axis to project into.
    a : float
        Distance to the origin for the plane to project into.
    v : list
        Normal vector of the plane to project into.
    
    Returns
    -------
    list
        The transformed vertices.
    
    """
    return [project_inv(pt, proj_axis, a, v) for pt in vertices]


def project_inv(pt, proj_axis, a, v):
    """Returns the vector w in the surface's plane such that project(w) equals
    x.
    
    See http://stackoverflow.com/a/39008641/1706564
    
    Parameters
    ----------
    pt : list 
        The two-dimensional point.
    proj_axis : int
        The axis to project into.
    a : float
        Distance to the origin for the plane to project into.
    v : list
        Normal vector of the plane to project into.
    
    Returns
    -------
    list
        The transformed point.
    
    """
    w = list(pt)
    w[proj_axis:proj_axis] = [0.0]
    c = a
    for i in range(3):
        c -= w[i] * v[i]
    c /= v[proj_axis]
    w[proj_axis] = c
    return tuple(w)


def pt_to_tuple(pt, dims=3):
    """Convert a point to a numpy array.
    
    Convert a Vector3D to an (x,y,z) tuple or a Vector2D to an (x,y) tuple.
    Ensures all values are floats since some other types cause problems in 
    pyclipper (notably where sympy.Zero is used to represent 0.0).

    Parameters
    ----------
    pt : sympy.Vector3D, sympy.Vector2D
        The point to convert.
    dims : int, optional
        Number of dimensions {default : 3}.
        
    Returns
    -------
    tuple

    """
    # handle Vector3D
    if dims == 3:
        return float(pt.x), float(pt.y), float(pt.z)
    # handle Vector2D
    elif dims == 2:
        return float(pt.x), float(pt.y)


def pt_to_array(pt, dims=3):
    """Convert a point to a numpy array.
    
    Converts a Vector3D to a numpy.array([x,y,z]) or a Vector2D to a 
    numpy.array([x,y]).
    Ensures all values are floats since some other types cause problems in 
    pyclipper (notably where sympy.Zero is used to represent 0.0).
    
    Parameters
    ----------
    pt : sympy.Vector3D
        The point to convert.
    dims : int, optional
        Number of dimensions {default : 3}.
    
    Returns
    -------
    numpy.ndarray

    """
    # handle Vector3D
    if dims == 3:
        return np.array([float(pt.x), float(pt.y), float(pt.z)])
    # handle Vector2D
    elif dims == 2:
        return np.array([float(pt.x), float(pt.y)])
    
    
def normalize_coords(poly, outside_pt, ggr=None):
    """Put coordinates into the correct format for EnergyPlus.
    
    poly : Polygon3D
        Polygon with the new coordinates.
    outside_pt : Vector3D
        An outside point of the new polygon.
    ggr : EPBunch, optional
        The section of the IDF that give rules for the order of vertices in a
        surface {default : None}.
    
    Returns
    -------
    list
    
    """
    # check and set entry direction
    poly = set_entry_direction(poly, outside_pt, ggr)
    # check and set starting position
    poly = set_starting_position(poly, outside_pt, ggr)

    return poly


def set_entry_direction(poly, outside_pt, ggr=None):
    """Check and set entry direction.
    """
    if not ggr:
        entry_direction = 'counterclockwise' # EnergyPlus default
    else:
        entry_direction = ggr.Vertex_Entry_Direction.lower()
    if entry_direction == 'counterclockwise':
        if poly.is_clockwise(outside_pt):
            poly = poly.invert_orientation()
    elif entry_direction == 'clockwise':
        if not poly.is_clockwise(outside_pt):
            poly = poly.invert_orientation()
    return poly


def set_starting_position(poly, outside_pt, ggr=None):
    """Check and set entry direction.
    """
    if not ggr:
        starting_position = 'upperleftcorner' # EnergyPlus default
    else:
        starting_position = ggr.Starting_Vertex_Position.lower()
    poly = poly.order_points(starting_position)

    return poly



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
Tests for Segment class, representing a line segment.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from geomeppy.segments import Segment
from geomeppy.vectors import Vector3D


def test_segment_repr():
    edge = Segment(Vector3D(0,0,0), Vector3D(2,0,0))
    assert eval(repr(edge)) == edge

def test_collinear():
    # same line
    edge1 = Segment(Vector3D(0,0,0), Vector3D(1,1,1))
    edge2 = Segment(Vector3D(0,0,0), Vector3D(1,1,1))
    assert edge1.is_collinear(edge2)

    # opposite direction line
    edge1 = Segment(Vector3D(1,1,1), Vector3D(0,0,0))
    edge2 = Segment(Vector3D(0,0,0), Vector3D(1,1,1))
    assert edge1.is_collinear(edge2)

    # edge1 is longer
    edge1 = Segment(Vector3D(0,0,0), Vector3D(4,4,4))
    edge2 = Segment(Vector3D(1,1,1), Vector3D(2,2,2))
    assert edge1.is_collinear(edge2)

    # same start point, different lengths
    edge1 = Segment(Vector3D(0,0,0), Vector3D(1,1,1))
    edge2 = Segment(Vector3D(0,0,0), Vector3D(2,2,2))
    assert edge1.is_collinear(edge2)
    
    # something being missed
    edge1 = Segment(Vector3D(1,4,0), Vector3D(1,0,0))
    edge2 = Segment(Vector3D(1,0,0), Vector3D(1,2,0))
    assert edge1.is_collinear(edge2)
    
    # parallel
    edge1 = Segment(Vector3D(0,0,0), Vector3D(1,1,1))
    edge2 = Segment(Vector3D(1,0,0), Vector3D(2,1,1))
    assert not edge1.is_collinear(edge2)




# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
Build IDF geometry from minimal inputs.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from geomeppy.polygons import Polygon3D
from geomeppy.vectors import Vector3D


class Zone(object):
    
    def __init__(self, name, surfaces):
        self.name = name
        self.walls = [s for s in surfaces['walls'] if s.area > 0]
        self.floors = surfaces['floors']
        self.roofs = surfaces['roofs']
        self.ceilings = surfaces['ceilings']
        

class Block(object):
    
    def __init__(self, name, coordinates, height, num_stories=1,
                 below_ground_stories=0, below_ground_storey_height=2.5):
        """Represents a single block for translation into an IDF.
        
        Parameters
        ----------
        name : str
            A name for the block.
        coordinates : list
            A list of (x, y) tuples representing the building outline.
        height : float
            The height of the block roof above ground level.
        num_stories : int, optional
            The total number of stories including basement stories. Default : 1.
        below_ground_stories : int, optional
            The number of stories below ground. Default : 0.
        below_ground_storey_height : float, optional
            The height of each basement storey. Default : 2.5.
        """
        self.name = name
        if coordinates[0] == coordinates[-1]:
            coordinates.pop()
        self.coordinates = coordinates
        self.height = height
        self.num_stories = num_stories
        self.num_below_ground_stories = below_ground_stories
        self.below_ground_storey_height = below_ground_storey_height
    
    @property
    def stories(self):
        """A list of dicts of the surfaces of each storey in the block.
        
        Returns
        -------
        list of dicts
            Dicts have the format:
                {'floors': [...], 'ceilings': [...],
                'walls': [...], 'roofs': [...]}
        
        """
        stories = []
        if self.num_below_ground_stories != 0:
            floor_no = -self.num_below_ground_stories
        else:
            floor_no = 0
        for f, c, w, r in zip(
                self.floors, self.ceilings, self.walls, self.roofs):
            stories.append({
                'storey_no': floor_no,
                'floors': f, 'ceilings': c, 'walls': w, 'roofs': r})
            floor_no += 1
        return stories
    
    @property
    def footprint(self):
        """Ground level outline of the block.
        
        Returns
        -------
        Polygon3D
        
        """
        coordinates = [(v[0], v[1], 0) for v in self.coordinates]
        return Polygon3D(coordinates)
    
    @property
    def storey_height(self):
        """Height of above ground stories.
        
        Returns
        -------
        float
        
        """
        return self.height / (self.num_stories - self.num_below_ground_stories)
    
    @property
    def floor_heights(self):
        """Floor height for each storey in the block.
        
        Returns
        -------
        list
        
        """
        lfl = self.lowest_floor_level
        sh = self.storey_height
        floor_heights = [lfl + sh * i for i in range(self.num_stories)]
        return floor_heights

    @property
    def ceiling_heights(self):
        """Ceiling height for each storey in the block.
        
        Returns
        -------
        list
        
        """
        lfl = self.lowest_floor_level
        sh = self.storey_height
        ceiling_heights = [lfl + sh * (i+1) for i in range(self.num_stories)]
        return ceiling_heights

    @property
    def lowest_floor_level(self):
        """Floor level of the lowest basement storey.
        
        Returns
        -------
        float
        
        """
        return -(self.num_below_ground_stories * 
                 self.below_ground_storey_height)
    
    @property
    def walls(self):
        """Coordinates for each wall in the block.
        
        These are ordered as a list of lists, one for each storey.
        
        Returns
        -------
        list
        
        """
        walls = []
        for fh, ch in zip(self.floor_heights, self.ceiling_heights):
            
            floor_walls = [make_wall(edge, fh, ch)
                           for edge in self.footprint.edges]
            walls.append(floor_walls)
        return walls
    
    @property
    def floors(self):
        """Coordinates for each floor in the block.
        
        Returns
        -------
        list
        
        """
        floors = [[self.footprint.invert_orientation() + Vector3D(0,0,fh)]
                  for fh in self.floor_heights]
        return floors

    @property
    def ceilings(self):
        """Coordinates for each ceiling in the block.
        
        Returns
        -------
        list
        
        """
        ceilings = [[self.footprint + Vector3D(0,0,ch)]
                    for ch in self.ceiling_heights[:-1]]
        
        ceilings.append('')
        return ceilings
    
    @property
    def roofs(self):
        """Coordinates for each roof of the block.
        
        This returns a list with an entry for each floor for consistency with
        the other properties of the Block object, but should only have roof
        coordinates in the list in the final position.
        
        Returns
        -------
        list
        
        """
        roofs = [[None] for ch in self.ceiling_heights[:-1]]
        roofs.append([self.footprint + Vector3D(0,0,self.height)])
        return roofs

    @property
    def surfaces(self):
        """Coordinates for all the surfaces in the block.
        
        Returns
        -------
        dict
        
        """
        return {'walls': self.walls, 
                'ceilings': self.ceilings,
                'roofs': self.roofs,
                'floors': self.floors}

def make_wall(edge, fh, ch):
    return Polygon3D([edge.p1 + (0,0,ch),  # upper left
                      edge.p1 + (0,0,fh),  # lower left
                      edge.p2 + (0,0,fh),  # lower right
                      edge.p2 + (0,0,ch),  # upper right
                      ])


# Copyright (c) 2016 Jamie Bull
# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
Monkey patches for changes to classes and functions in Eppy. These include
fixes which have not yet made it to the released version of Eppy. These will be
removed as soon as Eppy catches up. It also includes the addition of features
specific to geometry editing which may or may not be included future versions
of Eppy.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
from geomeppy.builder import Block
from geomeppy.builder import Zone
from geomeppy.intersect_match import getidfsubsurfaces
from geomeppy.intersect_match import getidfsurfaces
from geomeppy.intersect_match import intersect_idf_surfaces
from geomeppy.intersect_match import match_idf_surfaces
from geomeppy.intersect_match import set_coords
from geomeppy.recipes import set_default_constructions
from geomeppy.recipes import set_wwr
from geomeppy.recipes import translate
from geomeppy.recipes import translate_to_origin
from geomeppy.view_geometry import view_idf

from eppy import bunchhelpers
from eppy import iddgaps
from eppy.EPlusInterfaceFunctions import readidf
from eppy.bunch_subclass import EpBunch as BaseBunch
from eppy.idf_msequence import Idf_MSequence
from eppy.idfreader import convertallfields
from eppy.idfreader import iddversiontuple
from eppy.idfreader import idfreader1
from eppy.idfreader import makeabunch
from eppy.modeleditor import IDDNotSetError
from eppy.modeleditor import IDF as BaseIDF
from eppy.modeleditor import addthisbunch
from eppy.modeleditor import namebunch
from eppy.modeleditor import newrawobject
from eppy.modeleditor import obj2bunch
from py._log import warning


class EpBunch(BaseBunch):
    """Monkey-patched EpBunch to add the setcoords function.
    """
     
    def __init__(self, *args, **kwargs):
        super(EpBunch, self).__init__(*args, **kwargs)

    def setcoords(self, poly, ggr=None):
        """Set the coordinates of a surface.
         
        Parameters
        ----------
        poly : Polygon3D or list
             Either a Polygon3D object of a list of (x,y,z) tuples.
        ggr : EpBunch, optional
            A GlobalGeometryRules IDF object. Defaults to None.
             
        """
        surfaces = [
            "BUILDINGSURFACE:DETAILED",
            "WALL:DETAILED",
            "ROOFCEILING:DETAILED",
            "FLOOR:DETAILED",
            "FENESTRATIONSURFACE:DETAILED",
            "SHADING:SITE:DETAILED",
            "SHADING:BUILDING:DETAILED",
            "SHADING:ZONE:DETAILED", ]
        if self.key.upper() in surfaces:
            set_coords(self, poly, ggr)
        else:
            raise AttributeError
             

def idfreader1(fname, iddfile, theidf, conv=True, commdct=None, block=None):
    """read idf file and return bunches"""
    versiontuple = iddversiontuple(iddfile)
    block, data, commdct, idd_index = readidf.readdatacommdct1(
        fname,
        iddfile=iddfile,
        commdct=commdct,
        block=block)
    if conv:
        convertallfields(data, commdct)
    # fill gaps in idd
    ddtt, dtls = data.dt, data.dtls
    if versiontuple < (8,):
        skiplist = ["TABLE:MULTIVARIABLELOOKUP"]
    else:
        skiplist = None
    nofirstfields = iddgaps.missingkeys_standard(
        commdct, dtls,
        skiplist=skiplist)
    iddgaps.missingkeys_nonstandard(block, commdct, dtls, nofirstfields)
    # bunchdt = makebunches(data, commdct)
    bunchdt = makebunches(data, commdct, theidf)

    return bunchdt, block, data, commdct, idd_index, versiontuple


def addthisbunch(bunchdt, data, commdct, thisbunch, theidf):
    """add a bunch to model.
    abunch usually comes from another idf file
    or it can be used to copy within the idf file"""
    key = thisbunch.key.upper()
    obj = copy.copy(thisbunch.obj)
    abunch = obj2bunch(data, commdct, obj)
    bunchdt[key].append(abunch)
    return abunch


def makebunches(data, commdct, theidf):
    """make bunches with data"""
    bunchdt = {}
    dt, dtls = data.dt, data.dtls
    for obj_i, key in enumerate(dtls):
        key = key.upper()
        objs = dt[key]
        list1 = []
        for obj in objs:
            bobj = makeabunch(commdct, obj, obj_i)
            list1.append(bobj)
        bunchdt[key] = Idf_MSequence(list1, objs, theidf)
    return bunchdt


def obj2bunch(data, commdct, obj):
    """make a new bunch object using the data object"""
    dtls = data.dtls
    key = obj[0].upper()
    key_i = dtls.index(key)
    abunch = makeabunch(commdct, obj, key_i)
    return abunch


def makeabunch(commdct, obj, obj_i):
    """make a bunch from the object"""
    objidd = commdct[obj_i]
    objfields = [comm.get('field') for comm in commdct[obj_i]]
    objfields[0] = ['key']
    objfields = [field[0] for field in objfields]
    obj_fields = [bunchhelpers.makefieldname(field) for field in objfields]
    bobj = EpBunch(obj, obj_fields, objidd)
    return bobj


class IDF(BaseIDF):
    """Monkey-patched IDF.
    
    Patched to add read (to add additional functionality) and to fix 
    copyidfobject and newidfobject.
    
    """
    
    def __init__(self, *args, **kwargs):
        super(IDF, self).__init__(*args, **kwargs)
        
    def intersect_match(self):
        """Intersect all surfaces in the IDF, then set boundary conditions.
        """
        self.intersect()
        self.match()
        
    def intersect(self):
        """Intersect all surfaces in the IDF.
        """
        intersect_idf_surfaces(self)
        
    def match(self):
        """Set boundary conditions for all surfaces in the IDF.
        """
        match_idf_surfaces(self)
    
    def translate_to_origin(self):
        """
        Move an IDF close to the origin so that it can be viewed in SketchUp.
        """
        translate_to_origin(self)
    
    def translate(self, vector):
        """Move the IDF in the direction given by a vector.
        
        Parameters
        ----------
        vector : Vector2D, Vector3D, (x,y) or (x,y,z) list-like
            Representation of a vector to translate by.
            
        """
        surfaces = self.getsurfaces()
        translate(surfaces, vector)
    
    def set_default_constructions(self):
        set_default_constructions(self)
    
    def getsurfaces(self, surface_type=None):
        """Return all surfaces in the IDF.
        
        Returns
        -------
        list
        
        """
        return getidfsurfaces(self, surface_type)
    
    def getsubsurfaces(self, surface_type=None):
        """Return all subsurfaces in the IDF.
        
        Returns
        -------
        list
        
        """
        return getidfsubsurfaces(self, surface_type)
    
    def set_wwr(self, wwr):
        """Add strip windows to all external walls.
        
        Parameters
        ----------
        wwr : float
            Window to wall ratio in the range 0-1.
            
        """
        set_wwr(self, wwr)
    
    def view_model(self):
        """Show a zoomable, rotatable representation of the IDF.
        """
        view_idf(idf_txt=self.idfstr())
        
    def add_block(self, *args, **kwargs):
        """Add a block to the IDF
        """
        block = Block(*args, **kwargs)
        zoning = kwargs.get('zoning', 'by_storey')
        if zoning == 'by_storey':
            zones = [Zone('Block %s Storey %i' %
                          (block.name, storey['storey_no']), storey) 
                        for storey in block.stories]
        else:
            raise ValueError('%s is not a valid zoning rule' % zoning)
        for zone in zones:
            self.add_zone(zone)
    
    def add_shading_block(self, *args, **kwargs):
        """Add a shading block to the IDF
        """
        block = Block(*args, **kwargs)
        for i, wall in enumerate(block.walls[0], 1):
            if wall.area <= 0:
                continue
            s = self.newidfobject(
                'SHADING:SITE:DETAILED',
                '%s_%s' % (block.name, i))
            try:
                s.setcoords(wall)
            except ZeroDivisionError:
                self.removeidfobject(s)
        
    def add_zone(self, zone):
        try:
            ggr = self.idfobjects['GLOBALGEOMETRYRULES'][0]
        except IndexError:
            ggr = None
        # add zone object
        self.newidfobject('ZONE', zone.name)
        # add wall objects
        for i, surface_coords in enumerate(zone.walls, 1):
            if not surface_coords:
                continue
            name = '%s Wall %04d' % (zone.name, i)
            s = self.newidfobject('BUILDINGSURFACE:DETAILED', name,
                    Surface_Type = 'wall',
                    Zone_Name = zone.name)
            s.setcoords(surface_coords, ggr)
        # add floor objects
        for i, surface_coords in enumerate(zone.floors, 1):
            if not surface_coords:
                continue
            name = '%s Floor %04d' % (zone.name, i)
            s = self.newidfobject('BUILDINGSURFACE:DETAILED', name,
                    Surface_Type = 'floor',
                    Zone_Name = zone.name)
            s.setcoords(surface_coords, ggr)
        # add ceiling objects
        for i, surface_coords in enumerate(zone.ceilings, 1):
            if not surface_coords:
                continue
            name = '%s Ceiling %04d' % (zone.name, i)
            s = self.newidfobject('BUILDINGSURFACE:DETAILED', name,
                    Surface_Type = 'ceiling',
                    Zone_Name = zone.name)
            s.setcoords(surface_coords, ggr)
        # add roof objects
        for i, surface_coords in enumerate(zone.roofs, 1):
            if not surface_coords:
                continue
            name = '%s Roof %04d' % (zone.name, i)
            s = self.newidfobject('BUILDINGSURFACE:DETAILED', name,
                    Surface_Type = 'roof',
                    Zone_Name = zone.name)
            s.setcoords(surface_coords, ggr)

    def read(self):
        """
        Read the IDF file and the IDD file. If the IDD file had already been
        read, it will not be read again.

        Read populates the following data structures:

        - idfobjects : list
        - model : list
        - idd_info : list
        - idd_index : dict

        """
        if self.getiddname() == None:
            errortxt = ("IDD file needed to read the idf file. "
                        "Set it using IDF.setiddname(iddfile)")
            raise IDDNotSetError(errortxt)
        readout = idfreader1(
            self.idfname, self.iddname, self,
            commdct=self.idd_info, block=self.block)
        self.idfobjects, block, self.model, idd_info, idd_index, versiontuple = readout
        self.__class__.setidd(idd_info, idd_index, block, versiontuple)
            

    def newidfobject(self, key, aname='', **kwargs):
        """
        Add a new idfobject to the model. If you don't specify a value for a
        field, the default value will be set.

        For example ::

            newidfobject("CONSTRUCTION")
            newidfobject("CONSTRUCTION",
                Name='Interior Ceiling_class',
                Outside_Layer='LW Concrete',
                Layer_2='soundmat')

        Parameters
        ----------
        key : str
            The type of IDF object. This must be in ALL_CAPS.
        aname : str, deprecated
            This parameter is not used. It is left there for backward 
            compatibility.
        **kwargs
            Keyword arguments in the format `field=value` used to set the value
            of fields in the IDF object when it is created. 

        Returns
        -------
        EpBunch object

        """
        obj = newrawobject(self.model, self.idd_info, key)
        abunch = obj2bunch(self.model, self.idd_info, obj)
        if aname:
            warning.warn("The aname parameter should no longer be used.")
            namebunch(abunch, aname)
        self.idfobjects[key].append(abunch)
        for k, v in kwargs.items():
            abunch[k] = v
        return abunch

    def copyidfobject(self, idfobject):
        """Add an IDF object to the IDF.
        
        This has been monkey-patched to add the return value.

        Parameters
        ----------
        idfobject : EpBunch object
            The IDF object to remove. This usually comes from another idf file,
            or it can be used to copy within this idf file.
        
        Returns
        -------
        EpBunch object

        """
        abunch = addthisbunch(self.idfobjects,
                             self.model,
                             self.idd_info,
                             idfobject,
                             self)
         
        return abunch



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
Segment class, representing a line segment.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from geomeppy.vectors import Vector3D
from geomeppy.utilities import almostequal


class Segment(object):
    """Line segment in 3D."""
    
    def __init__(self, *vertices):
        self.vertices = vertices
        self.p1 = vertices[0]
        self.p2 = vertices[1]
    
    def __repr__(self):
        class_name = type(self).__name__
        return '{}({}, {})'.format(class_name, self.p1, self.p2)
    
    def __neg__(self):
        return Segment(self.p2, self.p1)
    
    def __iter__(self):
        return (i for i in self.vertices)
    
    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    
    def is_collinear(self, other):
        if almostequal(other, self) or almostequal(other, -self):
            return True
        a = self.p1 - other.p1
        b = self.p1 - other.p2
        angle_between = a.cross(b)
        if almostequal(angle_between, Vector3D(0,0,0)):
            return True
        a = self.p2 - other.p1
        b = self.p2 - other.p2
        angle_between = a.cross(b)
        if almostequal(angle_between, Vector3D(0,0,0)):
            return True
        return False
            
    def on_poly_edge(self, poly):
        """Test if segment lies on any edge of a polygon
        
        Parameters
        ----------
        poly : Polygon3D
            The polygon to test against.
        
        Returns
        -------
        bool
        
        """
        for edge in poly.edges:
            if self.is_collinear(edge):
                return True
        return False



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""pytest for recipes.py"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from geomeppy.intersect_match import getidfsurfaces
from geomeppy.intersect_match import intersect_idf_surfaces
from geomeppy.intersect_match import match_idf_surfaces
from geomeppy.polygons import Polygon3D
from geomeppy.recipes import set_wwr
from geomeppy.recipes import translate
from geomeppy.recipes import translate_to_origin
from geomeppy.utilities import almostequal

from eppy.iddcurrent import iddcurrent
from geomeppy.eppy_patches import IDF
from geomeppy.vectors import Vector2D
from geomeppy.vectors import Vector3D
from six import StringIO


idf_txt = """
Version, 8.5;
Building, Building 1, , , , , , , ;
Zone, z1 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
Zone, z2 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
BuildingSurface:Detailed, z1_FLOOR, Floor, , z1 Thermal Zone, ground, , NoSun, NoWind, , , 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0;
BuildingSurface:Detailed, z1_ROOF, Roof, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 2.0, 0.5, 1.0, 2.1, 0.5, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0001, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.0, 1.1, 0.5, 1.0, 1.1, 0.0, 1.0, 2.1, 0.0, 1.0, 2.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0002, Wall, , z1 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 1.0, 2.1, 0.5, 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 2.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0003, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 2.0, 0.5, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 2.0, 1.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0004, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z2_FLOOR, Floor, , z2 Thermal Zone, ground, , NoSun, NoWind, , , 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0;
BuildingSurface:Detailed, z2_ROOF, Roof, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 2.95, 0.5, 1.5, 3.05, 0.5, 1.5, 2.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0001, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 2.05, 0.5, 1.5, 2.05, 0.0, 1.5, 3.05, 0.0, 1.5, 3.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0002, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 3.05, 0.5, 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 2.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0003, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 2.95, 0.5, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 2.5, 1.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0004, Wall, , z2 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0, 1.5, 2.05, 0.5;
"""


class TestTranslate():
    
    def setup(self):
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        
        self.idf = IDF(StringIO(idf_txt))
            
    def test_translate(self):
        idf = self.idf
        surfaces = getidfsurfaces(idf)
        translate(surfaces, (50, 100))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected
        translate(surfaces, (-50, -100))  # move back
        
        translate(surfaces, Vector2D(50, 100))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected
        translate(surfaces, (-50, -100))  # move back
        
        translate(surfaces, Vector3D(50, 100, 0))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected


    def test_translate_to_origin(self):
        idf = self.idf
        surfaces = getidfsurfaces(idf)
        translate(surfaces, (50000, 10000))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [50002.0, 50002.0, 50001.0, 50001.0]
        assert result == expected
        
        translate_to_origin(idf)
        surfaces = getidfsurfaces(idf)
        min_x = min(min(Polygon3D(s.coords).xs) for s in surfaces)
        min_y = min(min(Polygon3D(s.coords).ys) for s in surfaces)
        assert min_x == 0
        assert min_y == 0


class TestMatchSurfaces():

    def setup(self):
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        self.idf = IDF(StringIO(idf_txt))
        intersect_idf_surfaces(self.idf)
        match_idf_surfaces(self.idf)
            
    def test_set_wwr(self):
        """Check that the correct WWR is set for all walls.
        """
        idf = self.idf
        wwr = 0.25
        set_wwr(idf, wwr)
        windows = idf.idfobjects['FENESTRATIONSURFACE:DETAILED']
        assert len(windows) == 8
        for window in windows:
            wall = idf.getobject('BUILDINGSURFACE:DETAILED',
                                 window.Building_Surface_Name)
            assert almostequal(window.area, wall.area * wwr, 3)



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""pytest for intersect_match.py"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from geomeppy.recipes import translate_coords
from geomeppy.view_geometry import view_polygons

from eppy.iddcurrent import iddcurrent
from geomeppy.eppy_patches import IDF
from geomeppy.intersect_match import get_adjacencies
from geomeppy.intersect_match import getidfsurfaces
from geomeppy.intersect_match import intersect
from geomeppy.intersect_match import intersect_idf_surfaces
from geomeppy.intersect_match import is_hole
from geomeppy.intersect_match import match_idf_surfaces
from geomeppy.intersect_match import unique
from geomeppy.polygons import Polygon3D
from geomeppy.utilities import almostequal
from six import StringIO


idf_txt = """
Version, 8.5;
Building, Building 1, , , , , , , ;
Zone, z1 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
Zone, z2 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
BuildingSurface:Detailed, z1_FLOOR, Floor, , z1 Thermal Zone, ground, , NoSun, NoWind, , , 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0;
BuildingSurface:Detailed, z1_ROOF, Roof, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 2.0, 0.5, 1.0, 2.1, 0.5, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0001, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.0, 1.1, 0.5, 1.0, 1.1, 0.0, 1.0, 2.1, 0.0, 1.0, 2.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0002, Wall, , z1 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 1.0, 2.1, 0.5, 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 2.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0003, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 2.0, 0.5, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 2.0, 1.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0004, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z2_FLOOR, Floor, , z2 Thermal Zone, ground, , NoSun, NoWind, , , 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0;
BuildingSurface:Detailed, z2_ROOF, Roof, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 2.95, 0.5, 1.5, 3.05, 0.5, 1.5, 2.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0001, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 2.05, 0.5, 1.5, 2.05, 0.0, 1.5, 3.05, 0.0, 1.5, 3.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0002, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 3.05, 0.5, 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 2.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0003, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 2.95, 0.5, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 2.5, 1.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0004, Wall, , z2 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0, 1.5, 2.05, 0.5;
"""

idf_txt_ring = """
Version, 8.5;
Building, Building 1, , , , , , , ;
Zone, Thermal Zone 1, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
Zone, Thermal Zone 2, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
BuildingSurface:Detailed, z1 Floor 0001, Floor, , Thermal Zone 2, Ground, , NoSun, NoWind, , , 0.0, 2.9, 0.0, 0.0, 0.0, 0.0, -2.14, 0.0, 0.0, -2.14, 2.9, 0.0;
BuildingSurface:Detailed, z1 Wall 0001, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , -2.14, 0.0, 1.5, -2.14, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.5;
BuildingSurface:Detailed, z1 Wall 0002, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , -2.14, 2.9, 1.5, -2.14, 2.9, 0.0, -2.14, 0.0, 0.0, -2.14, 0.0, 1.5;
BuildingSurface:Detailed, z1 Wall 0003, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , 0.0, 2.9, 1.5, 0.0, 2.9, 0.0, -2.14, 2.9, 0.0, -2.14, 2.9, 1.5;
BuildingSurface:Detailed, z1 Wall 0004, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , 0.0, 0.0, 1.5, 0.0, 0.0, 0.0, 0.0, 2.9, 0.0, 0.0, 2.9, 1.5;
BuildingSurface:Detailed, z1 Roof 0001, Roof, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , 0.0, 0.0, 1.5, 0.0, 2.9, 1.5, -2.14, 2.9, 1.5, -2.14, 0.0, 1.5;

BuildingSurface:Detailed, z2 Floor 0001, Floor, , Thermal Zone 1, Ground, , NoSun, NoWind, , , -0.259, 2.46, 1.5, -0.259, 0.4, 1.5, -1.68, 0.4, 1.5, -1.68, 2.46, 1.5;
BuildingSurface:Detailed, z2 Wall 0001, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -0.259, 2.46, 3.0, -0.259, 2.46, 1.5, -1.68, 2.46, 1.5, -1.68, 2.46, 3.0;
BuildingSurface:Detailed, z2 Wall 0002, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -0.259, 0.4, 3.0, -0.259, 0.4, 1.5, -0.259, 2.46, 1.5, -0.259, 2.46, 3.0;
BuildingSurface:Detailed, z2 Wall 0003, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -1.68, 0.4, 3.0, -1.68, 0.4, 1.5, -0.259, 0.4, 1.5, -0.259, 0.4, 3.0;
BuildingSurface:Detailed, z2 Wall 0004, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -1.68, 2.46, 3.0, -1.68, 2.46, 1.5, -1.68, 0.4, 1.5, -1.68, 0.4, 3.0;
BuildingSurface:Detailed, z2 Roof 0001, Roof, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -0.259, 0.4, 3.0, -0.259, 2.46, 3.0, -1.68, 2.46, 3.0, -1.68, 0.4, 3.0;
"""

class TestSetCoords():

    def setup(self):
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        
        self.idf = IDF(StringIO(idf_txt))
            
    def test_set_coords(self):
        idf = self.idf
        ggr = idf.idfobjects['GLOBALGEOMETRYRULES']        
        wall = idf.idfobjects['BUILDINGSURFACE:DETAILED'][0]
        poly1 = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        wall.setcoords(poly1, ggr)

def test_unique():
    poly1 = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
    poly2 = Polygon3D([(1,1,0),(1,0,0),(0,0,0),(0,1,0)])
    poly3 = Polygon3D([(0,1,0),(1,1,0),(1,0,0),(0,0,0)])
    assert poly1 != poly2
    assert poly2 == poly3
    polys = [poly1, poly2, poly3]
    unique_polys = unique(polys)
    assert len(unique_polys) == 2
    for poly in polys:
        assert poly in unique_polys
    
    
class TestSimpleTestPolygons():
    
    def test_simple_match(self):
        """
        The intersect function should just return the two polygons.
         ___  
        | 1 |
        |_2_| 

        """
        poly1 = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        poly2 = Polygon3D([(1,1,0),(1,0,0),(0,0,0),(0,1,0)])
        adjacencies = [(poly1, poly2)]
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result
    
    def test_simple_overlap(self):
        """
        The intersect function should return four surfaces.
         __ ___ __ 
        | 1| 1 |  |  
        |__|_2_|_2| 

         __ ___ __ 
        | 1| 3 |  |  
        |__|_4_|_2| 

        """
        poly1 = Polygon3D([(0,1,0),(0,0,0),(2,0,0),(2,1,0)])
        poly2 = Polygon3D([(3,1,0),(3,0,0),(1,0,0),(1,1,0)])
        adjacencies = [(poly1, poly2)]
        poly1 = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        poly2 = Polygon3D([(3,1,0),(3,0,0),(2,0,0),(2,1,0)])
        poly3 = Polygon3D([(1,1,0),(1,0,0),(2,0,0),(2,1,0)])
        poly4 = Polygon3D([(2,1,0),(2,0,0),(1,0,0),(1,1,0)])
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result

    def test_simple_hole(self):
        """
         _________ 
        | 1 ___   |  
        |  | 2 |  |
        |  |___|  | 
        |_________|
        
         ________ 
        |\ ___   |  
        |1| 2 | 4|
        | |_3_|  | 
        |/_______|
        
        """
        poly1 = Polygon3D([(0,4,0),(0,0,0),(4,0,0),(4,4,0)])
        poly2 = Polygon3D([(2,2,0),(2,1,0),(1,1,0),(1,2,0)])
        adjacencies = [(poly1, poly2)]
        
        poly1 = Polygon3D([(0,4,0), (0,0,0), (1,1,0), (1,2,0)])  # smaller section
        poly2 = Polygon3D([(1,2,0),(1,1,0),(2,1,0),(2,2,0)])  # inverse hole
        poly3 = Polygon3D([(2,2,0),(2,1,0),(1,1,0),(1,2,0)])  # hole
        poly4 = Polygon3D([(4,4,0), (0,4,0), (1,2,0), (2,2,0),
                           (2,1,0), (1,1,0), (0,0,0), (4,0,0)])  # larger section
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result
                
    def test_three_overlapping(self):
        """
         __ ___ __ __
        | 1| 1 | 3| 3|  
        |__|_2_|_2|__| 

         __ ___ __ __
        | 1| 2 | 4| 6|  
        |__|_3_|_5|__| 
        
        """
        poly1 = Polygon3D([(0,1,0),(0,0,0),(2,0,0),(2,1,0)])
        poly2 = Polygon3D([(3,1,0),(3,0,0),(1,0,0),(1,1,0)])
        poly3 = Polygon3D([(2,1,0),(2,0,0),(4,0,0),(4,1,0)])
        adjacencies = [(poly1, poly2),
                       (poly2, poly3)]
        poly1 = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        poly2 = Polygon3D([(2,1,0),(2,0,0),(1,0,0),(1,1,0)])
        poly3 = Polygon3D([(1,1,0),(1,0,0),(2,0,0),(2,1,0)])
        poly4 = Polygon3D([(2,1,0),(2,0,0),(3,0,0),(3,1,0)])
        poly5 = Polygon3D([(3,1,0),(3,0,0),(2,0,0),(2,1,0)])
        poly6 = Polygon3D([(3,1,0),(3,0,0),(4,0,0),(4,1,0)])
        expected = [poly1, poly2, poly3, poly4, poly5, poly6]
        
        result = intersect(*adjacencies[0])
        result.extend(intersect(*adjacencies[1]))
        result = unique(result)

        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result
    
    def test_double_overlap(self):
        """
         __________
        |__1_______| 
        | 1 | 2 |1 |
        |_2_|   |2_|
        |__________|
        
         __________
        |__1________| 
        | 2 | 4 | 5 |
        |_3_|   |_6_|
        |___________|
        
        """
        poly1 = Polygon3D([(0,2,0),(0,0,0),(3,0,0),(3,2,0)])
        poly2 = Polygon3D([(3,3,0),(3,1,0),(2,1,0),(2,2,0),
                           (1,2,0),(1,1,0),(0,1,0),(0,3,0)])
        adjacencies = [(poly1, poly2)]
        
        poly1 = Polygon3D([(3,3,0), (3,2,0), (0,2,0), (0,3,0)]) 
        poly2 = Polygon3D([(0,2,0), (0,1,0), (1,1,0), (1,2,0)])
        poly3 = Polygon3D([(1,2,0), (1,1,0), (0,1,0), (0,2,0)])
        poly4 = Polygon3D([(1,2,0), (1,1,0), (0,1,0), (0,0,0),
                           (3,0,0), (3,1,0), (2,1,0), (2,2,0)])
        poly5 = Polygon3D([(2,2,0), (2,1,0), (3,1,0), (3,2,0)])
        poly6 = Polygon3D([(3,2,0), (3,1,0), (2,1,0), (2,2,0)])

        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4, poly5, poly6]
        assert len(result) == len(expected)
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result
        
    def test_vertically_offset(self):
        """
         ___
        |_1_|
        | 1 |
        |_2_|
        |_2_|
         ___
        |_1_|
        | 2 |
        |_3_|
        |_4_|
         
        """
        poly1 = Polygon3D([(0,0,1),(0,0,0),(2,0,0),(2,0,1)])
        poly2 = Polygon3D([(3,0,1),(3,0,0),(1,0,0),(1,0,1)])
        adjacencies = [(poly1, poly2)]
        poly1 = Polygon3D([(0,0,1),(0,0,0),(1,0,0),(1,0,1)])
        poly2 = Polygon3D([(3,0,1),(3,0,0),(2,0,0),(2,0,1)])
        poly3 = Polygon3D([(1,0,1),(1,0,0),(2,0,0),(2,0,1)])
        poly4 = Polygon3D([(2,0,1),(2,0,0),(1,0,0),(1,0,1)])
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result

        
class TestMatchSurfaces():

    def setup(self):
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        
        self.idf = IDF(StringIO(idf_txt))
            
    def test_match_idf_surfaces(self):
        idf = self.idf
        intersect_idf_surfaces(idf)
        match_idf_surfaces(idf)
        idf.printidf()
        inside_wall = idf.getobject(
            'BUILDINGSURFACE:DETAILED', 'z1_WALL_0002_1')
        assert inside_wall.Outside_Boundary_Condition == 'surface'
        assert inside_wall.Outside_Boundary_Condition_Object == 'z2_WALL_0004_1'
        
        outside_wall = idf.getobject(
            'BUILDINGSURFACE:DETAILED', 'z1_WALL_0002_2')
        assert outside_wall.Outside_Boundary_Condition == 'outdoors'
        assert outside_wall.Outside_Boundary_Condition_Object == ''
        
        ground = idf.getobject(
            'BUILDINGSURFACE:DETAILED', 'z1_FLOOR')
        assert ground.Outside_Boundary_Condition == 'ground'
        assert ground.Outside_Boundary_Condition_Object == ''
        

class TestAdjacencies():
    
    def setup(self):
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        
        self.idf = IDF(StringIO(idf_txt))
            
    def test_get_adjacencies(self):
        surfaces = getidfsurfaces(self.idf)
        adjacencies = get_adjacencies(surfaces)
        assert (u'BuildingSurface:Detailed', u'z1_WALL_0002') in adjacencies
        assert (u'BuildingSurface:Detailed', u'z2_WALL_0004') in adjacencies
        assert len(adjacencies) == 2
        
        
def test_intersect():
    poly1 = Polygon3D([(1.0, 2.1, 0.5), (1.0, 2.1, 0.0),
                       (2.0, 2.0, 0.0), (2.0, 2.0, 0.5)])
    poly2 = Polygon3D([(2.5, 1.95, 0.5), (2.5, 1.95, 0.0),
                       (1.5, 2.05, 0.0), (1.5, 2.05, 0.5)])
    intersection = poly1.intersect(poly2)[0]
#    view_polygons({'blue': [poly1, poly2], 'red': [intersect]})
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)


def test_real_intersect():
    """
    Test that we can make a previously failing test pass by moving to the
    origin first.
    
    """
    poly1 = Polygon3D(
        [(526492.65, 185910.65, 6.0), (526492.65, 185910.65, 3.0),
         (526489.05, 185916.45, 3.0), (526489.05, 185916.45, 6.0)])
    poly2 = Polygon3D(
        [(526489.05, 185916.45, 5.0), (526489.05, 185916.45, 2.5),
        (526492.65, 185910.65, 2.5), (526492.65, 185910.65, 5.0)])
    min_x = min(min(s.xs) for s in [poly1, poly2])
    min_y = min(min(s.ys) for s in [poly1, poly2])
    poly1 = Polygon3D(translate_coords(poly1, [-min_x, -min_y, 0]))
    poly2 = Polygon3D(translate_coords(poly2, [-min_x, -min_y, 0]))
    intersection = Polygon3D(translate_coords(
        poly1.intersect(poly2)[0], [min_x, min_y, 0]))
    poly1 = Polygon3D(translate_coords(poly1, [min_x, min_y, 0]))
    poly2 = Polygon3D(translate_coords(poly2, [min_x, min_y, 0]))
#    view_polygons({'blue': [poly1, poly2]})#, 'red': [intersection]})
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)


def test_is_hole():
    """Test if a surface represents a hole in one of the surfaces.
    """
    # opposite faces (all touching edges)
    poly1 = Polygon3D([(0,4,0),(0,0,0),(4,0,0),(4,4,0)])
    poly2 = Polygon3D(reversed([(0,4,0),(0,0,0),(4,0,0),(4,4,0)]))
    intersection = poly1.intersect(poly2)[0]
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)

    # poly2 is within poly1 and reversed (no touching edges)
    poly1 = Polygon3D([(0,4,0),(0,0,0),(4,0,0),(4,4,0)])
    poly2 = Polygon3D(reversed([(1,3,0),(1,1,0),(3,1,0),(3,3,0)]))
    intersection = poly1.intersect(poly2)[0]
    assert is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)

    # poly2 is within poly1 and reversed (touches at x=0)
    poly1 = Polygon3D([(0,4,0),(0,0,0),(4,0,0),(4,4,0)])
    poly2 = Polygon3D(reversed([(0,3,0),(0,1,0),(3,1,0),(3,3,0)]))
    intersection = poly1.intersect(poly2)[0]
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)

    # poly2 overlaps poly1
    poly1 = Polygon3D([(1,4,0),(1,0,0),(5,0,0),(5,4,0)])
    poly2 = Polygon3D(reversed([(0,3,0),(0,1,0),(3,1,0),(3,3,0)]))
    intersection = poly1.intersect(poly2)[0]
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)


class TestIntersectMatchRing():
    
    def setup(self):
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        
        self.idf = IDF(StringIO(idf_txt_ring))
        
    def test_intersect_idf_surfaces(self):       
        idf = self.idf        
        starting = len(idf.idfobjects['BUILDINGSURFACE:DETAILED'])
        intersect_idf_surfaces(idf)
        idf.set_default_constructions()
        ending = len(idf.idfobjects['BUILDINGSURFACE:DETAILED'])
        assert starting == 12
        assert ending == 14
        for name in ['z1 Roof 0001_1',
                     'z1 Roof 0001_2',
                     'z1 Roof 0001_3',
                     'z2 Floor 0001_1']:
            obj = idf.getobject('BUILDINGSURFACE:DETAILED', name)
            assert obj


class TestIntersectMatch():
    
    def setup(self):
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        
        self.idf = IDF(StringIO(idf_txt))
            
    def test_getidfsurfaces(self):    
        idf = self.idf
        surfaces = getidfsurfaces(idf)
        assert len(surfaces) == 12
    
    def test_intersect_idf_surfaces(self):       
        idf = self.idf        
        starting = len(idf.idfobjects['BUILDINGSURFACE:DETAILED'])
        intersect_idf_surfaces(idf)
        ending = len(idf.idfobjects['BUILDINGSURFACE:DETAILED'])
        assert starting == 12
        assert ending == 14
        
        for name in ['z1_WALL_0002_1',
                     'z1_WALL_0002_2',
                     'z2_WALL_0004_1',
                     'z2_WALL_0004_2']:
            obj = idf.getobject('BUILDINGSURFACE:DETAILED', name)
            assert obj

def test_real_scale():
    """Test building, intersecting and matching from a real building footprint.
    """
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)
    
    idf = IDF(StringIO('Version, 8.5;'))
    poly1 = [(526492.65,185910.65),(526489.05,185916.45),
             (526479.15,185910.3),(526482.65,185904.6),
             (526492.65,185910.65)]
    poly2 = [(526483.3,185903.15),(526483.5,185903.25),
             (526482.65,185904.6),(526479.15,185910.3),
             (526489.05,185916.45),(526492.65,185910.65),
             (526493.4,185909.4),(526500,185913.95),
             (526500.45,185914.3),(526500,185914.85),
             (526497.4,185918.95),(526499.45,185920.2),
             (526494.4,185928.35),(526466.05,185910.95),
             (526471.1,185902.75),(526473.05,185903.9),
             (526476.2,185898.8),(526479.95,185901.1),
             (526483.3,185903.15)]
    idf.add_block('small', poly1, 6.0, 2)
    idf.add_block('large', poly2, 5.0, 2)
    idf.translate_to_origin()
    idf.intersect_match()
    idf.set_wwr(0.25)
    walls = idf.getsurfaces('wall')
    # look for a wall which should have been split
    assert 'Block large Storey 1 Wall 0003' not in [w.Name for w in walls]
    # look for another wall which should have been split
    assert 'Block large Storey 1 Wall 0005' not in [w.Name for w in walls]
    # look for a wall which should be an internal wall
    wall = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block small Storey 1 Wall 0002_1')
    assert wall.Outside_Boundary_Condition != 'outdoors'
    # look for another wall which should be an internal wall
    wall = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block large Storey 1 Wall 0003_2')
    assert wall.Outside_Boundary_Condition != 'outdoors'
    
    # look for two walls which are being incorrectly duplicated
    wall_1 = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block small Storey 0 Wall 0001_1')
    wall_2 = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block small Storey 0 Wall 0001_4')
    
    if wall_1 and wall_2:
        assert not almostequal(wall_1.coords, wall_2.coords)
    
    # look for two walls which are being incorrectly duplicated
    wall_1 = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block large Storey 1 Wall 0005_3')
    wall_2 = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block large Storey 1 Wall 0005_2')
    
    if wall_1 and wall_2:
        assert not almostequal(wall_1.coords, wall_2.coords)
    
    # look for two walls which are being incorrectly duplicated
    wall_1 = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block large Storey 1 Wall 0004_3')
    wall_2 = idf.getobject(
        'BUILDINGSURFACE:DETAILED', 'Block large Storey 1 Wall 0004_1')
    
    if wall_1 and wall_2:
        assert not almostequal(wall_1.coords, wall_2.coords)



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Intersect and match all surfaces in an IDF.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import defaultdict
from geomeppy.utilities import almostequal
from itertools import combinations
from itertools import product

from geomeppy.polygons import Polygon3D
from geomeppy.polygons import break_polygons
from geomeppy.vectors import Vector3D

from six.moves import zip_longest


def getidfplanes(surfaces):
    """Fast access data structure for potentially matched surfaces.
    
    Get a data structure populated with all the surfaces in the IDF, keyed by
    their distance from the origin, and their normal vector.
    
    Parameters
    ----------
    surfaces : list
        List of all the surfaces.
    
    Returns
    -------
    dict
        
    """
    planes = {}
    for s in surfaces:
        poly = Polygon3D(s.coords)
        rounded_distance = round(poly.distance, 8) 
        rounded_normal_vector = Vector3D(*[round(axis, 8)
                                           for axis in poly.normal_vector])
        planes.setdefault(rounded_distance, 
                          {}).setdefault(rounded_normal_vector, 
                                         []).append(s)
    return planes

def match_idf_surfaces(idf):
    """Match all surfaces in an IDF.
    
    Parameters
    ----------
    idf : IDF object
        The IDF.
    
    """
    surfaces = getidfsurfaces(idf)
    planes = getidfplanes(surfaces)
    for distance in planes:
        for vector in planes[distance]:
            surfaces = planes[distance][vector]
            matches = planes.get(-distance, {}).get(-vector, [])
            if not matches:
                # default set all the surfaces boundary conditions
                for s in surfaces:
                    set_unmatched_surface(s, vector)
            else:
                # check which are matches
                for s in surfaces:
                    for m in matches:
                        matched = False
                        poly1 = Polygon3D(s.coords)
                        poly2 = Polygon3D(m.coords)
                        if almostequal(poly1, poly2.invert_orientation()):
                            matched = True
                            # matched surfaces
                            s.Outside_Boundary_Condition = 'surface'
                            s.Outside_Boundary_Condition_Object = m.Name
                            s.Sun_Exposure = 'NoSun'
                            s.Wind_Exposure = 'NoWind'
                            # matched surfaces
                            m.Outside_Boundary_Condition = 'surface'
                            m.Outside_Boundary_Condition_Object = s.Name
                            m.Sun_Exposure = 'NoSun'
                            m.Wind_Exposure = 'NoWind'
                            break
                    if not matched:
                        # unmatched surfaces
                        set_unmatched_surface(s, vector)
                        set_unmatched_surface(m, vector)


def set_unmatched_surface(s, vector):
    """Set boundary conditions for a surface which does not adjoin another one.
    
    Parameters
    ----------
    s : EpBunch
        The surface.
    vector : Vector3D
        The surface normal vector
        
    """
    s.View_Factor_to_Ground = 'autocalculate'
    poly = Polygon3D(s.coords)
    if min(poly.zs) < 0 or all(z == 0 for z in poly.zs):
        # below ground or ground-adjacent surfaces
        s.Outside_Boundary_Condition_Object = ''
        s.Outside_Boundary_Condition = 'ground'
        s.Sun_Exposure = 'NoSun'
        s.Wind_Exposure = 'NoWind'
    else:
        s.Outside_Boundary_Condition = 'outdoors'
        s.Outside_Boundary_Condition_Object = ''
        s.Wind_Exposure = 'WindExposed'
        if almostequal(vector, (0, 0, -1)):
            # downward facing surfaces
            s.Sun_Exposure = 'NoSun'
        else:
            s.Sun_Exposure = 'SunExposed' # other external surfaces

            
def intersect_idf_surfaces(idf):
    """Intersect all surfaces in an IDF.
    
    Parameters
    ----------
    idf : IDF object
        The IDF.
    
    """
    surfaces = getidfsurfaces(idf)
    try:
        ggr = idf.idfobjects['GLOBALGEOMETRYRULES'][0]
    except IndexError:
        ggr = None
    # get all the intersected surfaces
    adjacencies = get_adjacencies(surfaces)
    for surface in adjacencies:
        key, name = surface
        new_surfaces = adjacencies[surface]
        old_obj = idf.getobject(key.upper(), name)
        for i, new_coords in enumerate(new_surfaces, 1):
            new = idf.copyidfobject(old_obj)
            new.Name = "%s_%i" % (name, i)
            set_coords(new, new_coords, ggr)
        idf.removeidfobject(old_obj)    


def get_adjacencies(surfaces):
    """Create a dictionary mapping surfaces to their adjacent surfaces.
    
    Parameters
    ----------
    surfaces : IdfMSequence
        A mutable list of surfaces.
    
    Returns
    -------
    dict
    
    """
    adjacencies = defaultdict(list)
    # find all adjacent surfaces
    for s1, s2 in combinations(surfaces, 2):
        adjacencies = populate_adjacencies(adjacencies, s1, s2)
    # make sure we have only unique surfaces
    for surface in adjacencies:
        adjacencies[surface] = unique(adjacencies[surface])

    return adjacencies


def populate_adjacencies(adjacencies, s1, s2):
    """Update the adjacencies dict with any intersections between two surfaces.
    
    Parameters
    ----------
    adjacencies : dict
        Dict to contain lists of adjacent surfaces.
    s1 : EPBunch
        Object representing an EnergyPlus surface.
    s2 : EPBunch
        Object representing an EnergyPlus surface.
    
    Returns
    -------
    dict
    
    """
    poly1 = Polygon3D(s1.coords)
    poly2 = Polygon3D(s2.coords)
    if not almostequal(abs(poly1.distance), abs(poly2.distance), 4):
        return adjacencies
    if not almostequal(poly1.normal_vector, poly2.normal_vector, 4):
        if not almostequal(poly1.normal_vector, -poly2.normal_vector, 4):
            return adjacencies
            
    intersection = poly1.intersect(poly2)
    if intersection:
        new_surfaces = intersect(poly1, poly2)
        new_s1 = [s for s in new_surfaces 
                  if almostequal(s.normal_vector, poly1.normal_vector, 4)]
        new_s2 = [s for s in new_surfaces 
                  if almostequal(s.normal_vector, poly2.normal_vector, 4)]
        adjacencies[(s1.key, s1.Name)] += new_s1
        adjacencies[(s2.key, s2.Name)] += new_s2
    return adjacencies


def intersect(poly1, poly2):
    """Calculate the polygons to represent the intersection of two polygons.
    
    Parameters
    ----------
    poly1 : Polygon3D
        The first polygon.
    poly2 : Polygon3D
        The second polygon.
    
    Returns
    -------
    list
        A list of unique polygons.
        
    """
    polys = []
    polys.extend(poly1.intersect(poly2))
    polys.extend(poly2.intersect(poly1))
    if is_hole(poly1, poly2):
        polys.extend(break_polygons(poly1, poly2))
    elif is_hole(poly2, poly1):
        polys.extend(break_polygons(poly2, poly1))
    else:
        polys.extend(poly1.difference(poly2))
        polys.extend(poly2.difference(poly1))
    polys = unique(polys)
    return polys


def unique(polys):
    """Make a unique set of polygons.
    
    Parameters
    ----------
    polys : list
    
    Returns
    -------
    list
    
    """
    flattened = []
    for item in polys:
        if isinstance(item, Polygon3D):
            flattened.append(item)
        elif isinstance(item, list):
            flattened.extend(item)

    results = []
    for poly in flattened:
        if not any(poly == result for result in results):
            results.append(poly)
    
    return results


def is_hole(surface, possible_hole):
    """Identify if an intersection is a hole in the surface.
    
    Check the intersection touches an edge of the surface. If it doesn't then
    it represents a hole, and this needs further processing into valid
    EnergyPlus surfaces.
    
    Parameters
    ----------
    surface : Polygon3D
        The first surface.

    possible_hole : Polygon3D
        The intersection into the surface.
        
    Returns
    -------
    bool

    """
    if surface.area < possible_hole.area:
        return False
    collinear_edges = (edges[0].is_collinear(edges[1]) 
                       for edges in product(surface.edges, possible_hole.edges))
    return not any(collinear_edges)


def set_coords(surface, coords, ggr):
    """Update the coordinates of a surface.
      
    Parameters
    ----------
    surface : EPBunch
        The surface to modify.
    coords : list
        The new coordinates as lists of [x,y,z] lists.
    ggr : EpBunch
        Global geometry rules.

    """
    poly = Polygon3D(coords)
    poly = poly.normalize_coords(ggr)
    coords = [i for vertex in poly for i in vertex]
    # find the vertex fields
    n_vertices_index = surface.objls.index('Number_of_Vertices')
    first_x = n_vertices_index + 1  # X of first coordinate
    surface.obj = surface.obj[:first_x]    
    # set the vertex field values
    surface.fieldvalues.extend(coords)


def getidfsurfaces(idf, surface_type=None):
    """Return all surfaces in an IDF.
    
    Parameters
    ----------
    idf : IDF object
        The IDF to search.
    surface_type : str, optional
        A surface type to specify. Default is None, which returns all surfaces.
        
    """
    surfaces = idf.idfobjects['BUILDINGSURFACE:DETAILED']
    if surface_type:
        surfaces = [s for s in surfaces if s.Surface_Type == surface_type]
    return surfaces


def getidfsubsurfaces(idf, surface_type=None):
    """Return all subsurfaces in an IDF.
    
    Parameters
    ----------
    idf : IDF object
        The IDF to search.
    surface_type : str, optional
        A surface type to specify. Default is None, which returns all surfaces.
        
    """
    surfaces = idf.idfobjects['FENESTRATIONSURFACE:DETAILED']
    if surface_type:
        surfaces = [s for s in surfaces 
                    if s.Surface_Type.lower() == surface_type.lower()]
    return surfaces



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""utilities for use in test suite"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

def almostequal(first, second, places=7):
    """Tests a range of types for near equality"""
    try:
        # try converting to float first
        first = float(first)
        second = float(second)
        # test floats for near-equality
        if round(abs(second - first), places) != 0:
            return False
        else:
            return True
    except ValueError:
        # handle non-float types
        return str(first) == str(second)
    except TypeError:
        # handle iterables
        return all([almostequal(a, b, places) for a,b in zip(first, second)])



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
A module to handle translations, using Christopher Gohlke's transforms3d as far
as possible, but also trying to respect the intent of the algorithms used in 
OpenStudio for the sake of consistency between tools based on EnergyPlus.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from transforms3d._gohlketransforms import concatenate_matrices
from transforms3d._gohlketransforms import identity_matrix
from transforms3d._gohlketransforms import inverse_matrix
from transforms3d._gohlketransforms import rotation_matrix
from transforms3d._gohlketransforms import translation_matrix

from geomeppy.vectors import Vector3D

import numpy as np


class Transformation(object):
        
    def __init__(self, mat=None):
        if mat is None:
            # initialise with a 4D identity matrix
            self.matrix = identity_matrix()
        else:
            # initialise with the matrix passed in
            self.matrix = mat
        assert self.matrix.shape == (4, 4)
        
    def __mul__(self, other):
        if hasattr(other, 'matrix'):
            # matrix by a matrix
            mat = concatenate_matrices(self.matrix, other.matrix)
            return Transformation(mat)
        elif hasattr(other, 'x'):
            # matrix by a vector
            vector = other
            temp = [vector.x, vector.y, vector.z, 1]
            result = np.dot(self.matrix, temp)[:3]
            return Vector3D(*result)
        else:
            # matrix by each point in a polygon
            result = [self * point for point in other]
            return other.__class__(result)
    
    def align_z_prime(self, zp):
        """Transform system with z' to regular system. Will try to align y' 
        with z, but if that fails will align y' with y
        
        """
        zp = zp.normalize()
        
        x_axis = Vector3D(1,0,0)
        y_axis = Vector3D(0,1,0)
        z_axis = Vector3D(0, 0, 1)
        neg_x_axis = Vector3D(-1, 0, 0)
        
        # check if face normal is up or down
        if abs(zp.dot(z_axis)) < 0.99:
            # not facing up or down, set yPrime along z_axis
            yp = z_axis - zp.dot(z_axis) * zp
            yp = yp.normalize()
            xp = yp.cross(zp)
        else:
            # facing up or down, set xPrime along -x_axis
            xp = neg_x_axis - zp.dot(neg_x_axis) * zp
            xp = xp.normalize()
            yp = zp.cross(xp)
        
        self.matrix[:3,0] = xp
        self.matrix[:3,1] = yp
        self.matrix[:3,2] = zp
        
        return self
        
    def align_face(self, polygon):
        """A transformation to align a polygon with the origin.
        
        Parameters
        ----------
        polygon : Polygon3D
        
        Returns
        -------
        Transformation
        
        """
        zp = polygon.normal_vector
        align = self.align_z_prime(zp)
        
        aligned_vertices = align.inverse() * polygon
        min_x = min(aligned_vertices.xs)
        min_y = min(aligned_vertices.ys)
        min_z = min(aligned_vertices.zs)
        assert aligned_vertices.is_horizontal
        
        direction = Vector3D(min_x, min_y, min_z)
        translate = self.translation(direction)
        
        self.matrix = concatenate_matrices(align.matrix, translate.matrix)

        return self
    
    def inverse(self):
        """The inverse of a given transformation.
        
        Returns
        -------
        Transformation
        
        """
        return Transformation(inverse_matrix(self.matrix))
    
    def translation(self, direction):
        return Transformation(translation_matrix(direction))

    def rotation(self, direction, angle):
        return Transformation(rotation_matrix(angle, direction))


def align_face(polygon):
    """Transformation to align face with z-axis.

    Parameters
    ----------
    polygon : Polygon3D
        Polygon to be aligned.
    
    Returns
    -------
    Polygon3D
    
    """
    t = Transformation()
    t.align_face(polygon)
    
    return t.inverse() * polygon


def invert_align_face(original, poly2):
    """Transformation to align face with original position.
    
    Parameters
    ----------
    original : Polygon3D
        Polygon in the desired orientation.
    poly2 : Polygon3D
        Polygon previously aligned with `align_face`.
    
    Returns
    -------
    Polygon3D
    
    """
    t = Transformation()
    t.align_face(original)
    
    return t * poly2



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Utilities for IDF vectors.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np


class Vector2D(object):
    """Two dimensional point."""
    def __init__(self, *args):
        self.args = list(args)
        self.x = float(args[0])
        self.y = float(args[1])
    
    def __iter__(self):
        return (i for i in self.args)

    def __repr__(self):
        class_name = type(self).__name__
        return '{}({!r}, {!r})'.format(class_name, *self.args)
    
    def __eq__(self, other):
        for a, b in zip(self, other):
            if a != b:
                return False
        return True
    
    def __sub__(self, other):
        return self.__class__(*[self[i] - other[i] for i in range(len(self))])
    
    def __add__(self, other):
        return self.__class__(*[self[i] + other[i] for i in range(len(self))])
    
    def __neg__(self):
        return self.__class__(*inverse_vector(self))
        
    def __len__(self):
        return len(self.args)
    
    def __getitem__(self, key):
        return self.args[key]

    def __setitem__(self, key, value):
        self.args[key] = value
    
    def __hash__(self):
        return hash(self.x) ^ hash(self.y)
        
    def dot(self, other):
        return np.dot(self, other)

    def cross(self, other):
        return np.cross(self, other)

    @property
    def length(self):
        """The length of a vector.
        
        Parameters
        ----------
        list-like
            A list or other iterable.
            
        """
        length = sum(x ** 2 for x in self) ** 0.5
    
        return length

    def closest(self, poly):
        """Find the closest vector in a polygon.
        
        Parameters
        ----------
        poly : Polygon or Polygon3D
        
        """
        min_d = float('inf')
        closest_pt = None
        for pt2 in poly:
            direction = self - pt2
            sq_d = sum(x ** 2 for x in direction)
            if sq_d < min_d:
                min_d = sq_d
                closest_pt = pt2
        return closest_pt
        
    def normalize(self):
        return self.set_length(1.0)
    
    def set_length(self, new_length):
        current_length = self.length        
        multiplier = new_length / current_length
        self.args = [i * multiplier for i in self.args]
        return self
    
    def invert(self):
        return -self

class Vector3D(Vector2D):
    """Three dimensional point."""
    def __init__(self, x, y, z=0):
        super(Vector3D, self).__init__(x, y)
        self.z = float(z)
        self.args = (self.x, self.y, self.z)
        
    def __repr__(self):
        class_name = type(self).__name__
        return '{}({!r}, {!r}, {!r})'.format(class_name, *self.args)
    
    def __hash__(self):
        return hash(self.x) ^ hash(self.y) ^ hash(self.z)
        

def normalise_vector(v):
    """Convert a vector to a unit vector
    
    Parameters
    ----------
    v : list
        The vector.
        
    Returns
    -------
    list
    
    """
    magnitude = sum(abs(i) for i in v)
    normalised_v = [i / magnitude for i in v]
    
    return normalised_v


def inverse_vector(v):
    """Convert a vector to the same vector but in the opposite direction
    
    Parameters
    ----------
    v : list
        The vector.
        
    Returns
    -------
    list
    
    """    
    return [-i for i in v]




# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""pytest for transformations.py"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from transforms3d._gohlketransforms import translation_matrix

from geomeppy.polygons import Polygon3D
from geomeppy.transformations import Transformation
from geomeppy.vectors import Vector3D
import numpy as np
from geomeppy.utilities import almostequal


class TestTransormations():
    
    def test_translation_transformations(self):
        tol = 12 # places
        trans = Vector3D(1,1,1)
        point1 = Vector3D(1,0,0)
        t = Transformation()
        
        # identity transformation
        temp = t * point1
        assert almostequal(1.0, temp.x, tol)
        assert almostequal(0.0, temp.y, tol)
        assert almostequal(0.0, temp.z, tol)
        
        # move by 1, 1, 1
        t = Transformation(translation_matrix(trans))
        temp = t * point1
        assert almostequal(2.0, temp.x, tol)
        assert almostequal(1.0, temp.y, tol)
        assert almostequal(1.0, temp.z, tol)
        
        # move by -1, -1, -1
        temp = t.inverse() * point1
        assert almostequal(0.0, temp.x, tol)
        assert almostequal(-1.0, temp.y, tol)
        assert almostequal(-1.0, temp.z, tol)
        
        # identity transformation
        temp = t.inverse() * t * point1
        assert almostequal(1.0, temp.x, tol)
        assert almostequal(0.0, temp.y, tol)
        assert almostequal(0.0, temp.z, tol)
        
        # identity transformation
        temp = t * t.inverse() * point1
        assert almostequal(1.0, temp.x, tol)
        assert almostequal(0.0, temp.y, tol)
        assert almostequal(0.0, temp.z, tol)
    
    def test_align_z_prime_transformations(self):
        x_axis = Vector3D(1,0,0)
        y_axis = Vector3D(0,1,0)
        z_axis = Vector3D(0,0,1)
        t = Transformation()
        
        outward_normal = Vector3D(0, -1, 0)
        t = t.align_z_prime(outward_normal)
        assert x_axis == t * x_axis
        assert z_axis == t * y_axis
        result = t * z_axis
        assert outward_normal == result
        
        outward_normal = Vector3D(1, 0, 0)
        t = t.align_z_prime(outward_normal)
        assert y_axis == t * x_axis
        assert z_axis == t * y_axis
        assert outward_normal == t * z_axis
        
        outward_normal = Vector3D(0, 1, 0)
        t = t.align_z_prime(outward_normal)
        assert -x_axis == t * x_axis
        assert z_axis == t * y_axis
        assert outward_normal == t * z_axis
        
        outward_normal = Vector3D(-1, 0, 0)
        t = t.align_z_prime(outward_normal)
        assert -y_axis == t * x_axis
        assert z_axis == t * y_axis
        assert outward_normal == t * z_axis
        
        outward_normal = Vector3D(0, 0, 1)
        t = t.align_z_prime(outward_normal)
        assert -x_axis == t * x_axis
        assert -y_axis == t * y_axis
        assert outward_normal == t * z_axis
        
        outward_normal = Vector3D(0, 0, -1)
        t = t.align_z_prime(outward_normal)
        assert -x_axis == t * x_axis
        assert y_axis == t * y_axis
        assert outward_normal == t * z_axis
    
    def test_align_face_transformations(self):
        tol = 12  # places
        
        vertices = Polygon3D([(1, 0, 1), (1, 0, 0), (2, 0, 0), (2, 0, 1)])
        t = Transformation()
        
        # rotate 0 degrees about z
        testVertices = t.rotation(Vector3D(0,0,1), 0) * vertices
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices
        expectedVertices = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        assert almostequal(tempVertices, expectedVertices, tol)
    
        # rotate 30 degrees about z
        testVertices = t.rotation(Vector3D(0,0,1), np.deg2rad(30)) * vertices
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices
        expectedVertices = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        assert almostequal(tempVertices, expectedVertices, tol)
    
        # rotate -30 degrees about z
        testVertices = t.rotation(Vector3D(0,0,1), -np.deg2rad(30)) * vertices
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices
        expectedVertices = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        assert almostequal(tempVertices, expectedVertices, tol)
    
        # rotate -30 degrees about x
        testVertices = t.rotation(Vector3D(1,0,0), -np.deg2rad(30)) * vertices
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices    
        expectedVertices = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        assert almostequal(tempVertices, expectedVertices, tol)
    
        # rotate -90 degrees about x
        testVertices = t.rotation(Vector3D(1,0,0), -np.deg2rad(90)) * vertices
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices
        expectedVertices = Polygon3D([(1,0,0),(1,1,0),(0,1,0),(0,0,0)])
        assert almostequal(tempVertices, expectedVertices, tol)
    
        # rotate 30 degrees about x
        testVertices = t.rotation(Vector3D(1,0,0), np.deg2rad(30)) * vertices
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices    
        expectedVertices = Polygon3D([(0,1,0),(0,0,0),(1,0,0),(1,1,0)])
        assert almostequal(tempVertices, expectedVertices, tol)
    
        # rotate 90 degrees about x
        testVertices = t.rotation(Vector3D(1,0,0), np.deg2rad(90)) * vertices
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices    
        expectedVertices = Polygon3D([(1,0,0),(1,1,0),(0,1,0),(0,0,0)])
        assert almostequal(tempVertices, expectedVertices, tol)
    
    
    def test_align_face_transformations_trapezoid_floor(self):
        tol = 12  # places
        
        testVertices = Polygon3D([(27.69,0,0),(0,0,0),
                                  (5,5,0),(22.69,5,0)])
        t = Transformation().align_face(testVertices)
        tempVertices = t.inverse() * testVertices
        expectedVertices = Polygon3D([(0,0,0),(27.69,0,0),(22.69,5,0),(5,5,0)])
        assert almostequal(tempVertices, expectedVertices, tol)



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
Recipes for making changes to EnergyPlus IDF files.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from geomeppy.intersect_match import getidfsurfaces
from geomeppy.polygons import Polygon3D

from geomeppy.vectors import Vector3D


def set_default_constructions(idf):
    constructions = ['Project Wall', 'Project Partition','Project Floor',
                     'Project Flat Roof', 'Project Ceiling',
                     'Project External Window', 'Project Door']
    for construction in constructions:
        idf.newidfobject('CONSTRUCTION', construction,
                         Outside_Layer='DefaultMaterial')
    idf.newidfobject('MATERIAL', 'DefaultMaterial',
                     Roughness='Rough',
                     Thickness=0.1,
                     Conductivity=0.1,
                     Density=1000,
                     Specific_Heat=1000,
                     )

    for surface in idf.getsurfaces():
        set_default_construction(surface)
    for subsurface in idf.getsubsurfaces():
        set_default_construction(subsurface)


def set_default_construction(surface):
    if surface.Surface_Type.lower() == 'wall':
        if surface.Outside_Boundary_Condition.lower() == 'outdoors':
            surface.Construction_Name = 'Project Wall'
        elif surface.Outside_Boundary_Condition.lower() == 'ground':
            surface.Construction_Name = 'Project Wall'
        else:
            surface.Construction_Name = 'Project Partition'
    if surface.Surface_Type.lower() == 'floor':
        if surface.Outside_Boundary_Condition.lower() == 'ground':
            surface.Construction_Name = 'Project Floor'
        else:
            surface.Construction_Name = 'Project Floor'
    if surface.Surface_Type.lower() == 'roof':
        surface.Construction_Name = 'Project Flat Roof'
    if surface.Surface_Type.lower() == 'ceiling':
        surface.Construction_Name = 'Project Ceiling'
    if surface.Surface_Type == 'window':
        surface.Construction_Name = 'Project External Window'
    if surface.Surface_Type == 'door':
        surface.Construction_Name = 'Project Door' 


def translate_to_origin(idf):
    """
    Move an IDF close to the origin so that it can be viewed in SketchUp.
    
    Parameters
    ----------
    idf : IDF object
    
    """
    surfaces = getidfsurfaces(idf)
    windows = idf.idfobjects['FENESTRATIONSURFACE:DETAILED']
        
    min_x = min(min(Polygon3D(s.coords).xs) for s in surfaces)
    min_y = min(min(Polygon3D(s.coords).ys) for s in surfaces)

    translate(surfaces, (-min_x, -min_y))
    translate(windows, (-min_x, -min_y))
    
    
def translate(surfaces, vector):
    """Translate all surfaces by a vector.
    
    Parameters
    ----------
    surfaces : list
        A list of EpBunch objects.
    vector : Vector2D, Vector3D, (x,y) or (x,y,z) list-like
        Representation of a vector to translate by.
        
    """
    vector = Vector3D(*vector)
    for s in surfaces:
        new_coords = translate_coords(s.coords, vector)
        s.setcoords(new_coords)


def translate_coords(coords, vector):
    """Translate a set of coords by a direction vector.
    
    Parameters
    ----------
    coords : list
        A list of points.
    vector : Vector2D, Vector3D, (x,y) or (x,y,z) list-like
        Representation of a vector to translate by.
    
    Returns
    -------
    list of Vector3D objects
    
    """
    return [Vector3D(*v) + vector for v in coords]


def set_wwr(idf, wwr=0.2):
    """Set the window to wall ratio on all external walls.
    
    Parameters
    ----------
    idf : IDF object
        The IDF to edit.
    wwr : float
        The window to wall ratio.
        
    """
    try:
        ggr = idf.idfobjects['GLOBALGEOMETRYRULES'][0]
    except IndexError:
        ggr = []
    walls = [s for s in idf.idfobjects['BUILDINGSURFACE:DETAILED']
             if s.Surface_Type.lower() == 'wall'
             and s.Outside_Boundary_Condition.lower() == 'outdoors']
    windows = idf.idfobjects['FENESTRATIONSURFACE:DETAILED']
    for window in windows:
        idf.removeidfobject(window)
    for wall in walls:
        coords = window_vertices_given_wall(wall, wwr)
        window = idf.newidfobject(
            'FENESTRATIONSURFACE:DETAILED',
            Name = "%s window" % wall.Name,
            Surface_Type = 'Window',
            Building_Surface_Name = wall.Name,
            View_Factor_to_Ground = 'autocalculate', # from the surface angle
            )
        window.setcoords(coords, ggr)
    
    
def window_vertices_given_wall(wall, wwr):
    """Calculate window vertices given wall vertices and glazing ratio.
    
    For each axis:
    1) Translate the axis points so that they are centred around zero
    2) Either:
        a) Multiply the z dimension by the glazing ratio to shrink it vertically
        b) Multiply the x or y dimension by 0.995 to keep inside the surface
    3) Translate the axis points back to their original positions
    
    Parameters
    ----------
    wall : EpBunch
        The wall to add a window on. We expect each wall to have four vertices.
    wwr : float
        Window to wall ratio.
    
    Returns
    -------
    list 
        Window vertices bounding a vertical strip midway up the surface.
    
    """
    vertices = wall.coords
    average_x = sum([x for x, _y, _z in vertices]) / len(vertices)
    average_y = sum([y for _x, y, _z in vertices]) / len(vertices)
    average_z = sum([z for _x, _y, z in vertices]) / len(vertices)
    # move windows in 0.5% from the edges so they can be drawn in SketchUp
    window_points = [[
                      ((x - average_x) * 0.999) + average_x,
                      ((y - average_y) * 0.999) + average_y,
                      ((z - average_z) * wwr) + average_z
                      ]
                     for x, y, z in vertices]

    return Polygon3D(window_points)
